//! Tag mod
//!
//! An event can tag/reference another public key, or another event.
//!
//! It can also use the tag to add all sort of metadata, useful for their own
//! type
use super::{filter::TagValue, Addr};
use serde::{
    de::{self, Deserialize, Deserializer},
    ser::{self, SerializeSeq, Serializer},
};
use std::collections::VecDeque;
use url::Url;

mod event;
mod public_key;

pub use self::{event::*, public_key::*};

#[derive(Debug, PartialEq, Eq, Clone)]
/// Relay access type
pub enum RelayAccessType {
    /// Read access
    Read,
    /// Write access
    Write,
    /// Both read and write access
    Both,
}

/// Tags are how events relates to each other.
///
/// So far, there are two standard tags, Events and Public Key. This an event
/// can be related to another, with extra options (like a relayer url, and a pet
/// name).
///
/// Any non standard tag will be parsed as Unknown with a vector of strings as
/// their parameters
#[derive(Debug, PartialEq, Eq, Clone)]
pub enum Tag {
    /// Tag another event
    Event(Event),
    /// Tag another public key
    PubKey(PubKey),
    /// Tag a relay
    Relay(Url, RelayAccessType),
    /// Tag a hashtag
    Hashtag(String),
    /// Tag with an external content id
    ExternalContentId(String, Option<Url>),
    /// Tag with a title
    Title(String),
    /// Any non standard tag
    Unknown(String, Vec<String>),
    /// Weird, supported nonetheless
    Empty,
}

impl Tag {
    /// Is the tag a public key?
    pub fn is_pubkey(&self) -> bool {
        matches!(self, Self::PubKey(_))
    }

    /// Is the tag an event?
    pub fn is_event(&self) -> bool {
        matches!(self, Self::Event(_))
    }

    /// Get the identifier for the tag
    pub fn get_identifier(&self) -> &str {
        match self {
            Tag::Event(_) => "e",
            Tag::PubKey(_) => "p",
            Tag::Relay(_, _) => "r",
            Tag::Hashtag(_) => "t",
            Tag::Title(_) => "title",
            Tag::ExternalContentId(_, _) => "i",
            Tag::Unknown(u, _) => u,
            Tag::Empty => "",
        }
    }

    /// Get the indexable value of the tag
    pub fn get_indexable_value(&self) -> Option<TagValue> {
        match self {
            Tag::Event(event) => Some(TagValue::Address(event.id.clone())),
            Tag::PubKey(key) => Some(TagValue::Address(key.id.clone())),
            Tag::ExternalContentId(id, _) => Some(TagValue::String(id.clone())),
            Tag::Hashtag(content) | Tag::Title(content) => Some(TagValue::String(content.clone())),
            Tag::Relay(url, _) => Some(TagValue::String(url.to_string())),
            Tag::Unknown(_, args) => args.get(0).cloned().map(TagValue::String),
            Tag::Empty => None,
        }
    }

    /// Converts the tag into bytes if possible.
    ///
    /// The bytes can be used to store the tag in a database or to find it in an in-memory index for subscribers
    pub fn into_bytes(self) -> Option<Vec<u8>> {
        Some(
            self.get_indexable_value()?
                .into_bytes(self.get_identifier()),
        )
    }
}

impl ser::Serialize for Tag {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut seq = serializer.serialize_seq(Some(2))?;
        seq.serialize_element(self.get_identifier())?;

        match self {
            Tag::Event(event) => {
                seq.serialize_element(&event.id.to_hex())?;
                if let Some(relayer) = &event.relayer_url {
                    seq.serialize_element(&relayer)?;
                    if let Some(marker) = &event.marker {
                        seq.serialize_element(&marker.to_string())?;
                    }
                }
            }
            Tag::ExternalContentId(content, url) => {
                seq.serialize_element(content)?;
                if let Some(url) = url {
                    seq.serialize_element(url)?;
                }
            }
            Tag::Hashtag(content) | Tag::Title(content) => {
                seq.serialize_element(content)?;
            }
            Tag::PubKey(key) => {
                seq.serialize_element(&key.id.to_hex())?;
                if let Some(relayer) = &key.relayer_url {
                    seq.serialize_element(relayer)?;
                    if let Some(pet_name) = &key.pet_name {
                        seq.serialize_element(pet_name)?;
                    }
                }
            }
            Tag::Relay(url, access) => {
                seq.serialize_element(url)?;

                if let Some(access) = match access {
                    RelayAccessType::Read => Some("read"),
                    RelayAccessType::Write => Some("write"),
                    RelayAccessType::Both => None,
                } {
                    seq.serialize_element(access)?;
                }
            }
            Tag::Unknown(_, extra) => {
                for extra in extra.iter() {
                    seq.serialize_element(extra)?;
                }
            }
            Tag::Empty => unreachable!(),
        }
        seq.end()
    }
}

impl<'de> Deserialize<'de> for Tag {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let mut parts: VecDeque<String> = Deserialize::deserialize(deserializer)?;

        match parts.len() {
            0 => return Ok(Tag::Empty),
            1 => {
                return Ok(Tag::Unknown(parts.pop_front().unwrap_or_default(), vec![]));
            }
            _ => {}
        }

        let tag_type = parts.pop_front().unwrap_or_default();
        let default = Tag::Unknown(tag_type.clone(), parts.clone().into());

        let tag: Result<_, D::Error> = match tag_type.as_str() {
            "e" | "p" => parts
                .pop_front()
                .ok_or_else::<D::Error, _>(|| de::Error::custom("missing argument"))
                .and_then(|id| {
                    Addr::try_from_public_key_str(&id, None)
                        .map(|id| {
                            let relayer_url = parts.pop_front();
                            let extra = parts.pop_front();

                            if "e" == tag_type {
                                Tag::Event(Event {
                                    id,
                                    relayer_url,
                                    marker: extra.map(|x| x.as_str().into()),
                                })
                            } else {
                                Tag::PubKey(PubKey {
                                    id,
                                    relayer_url,
                                    pet_name: extra,
                                })
                            }
                        })
                        .map_err(de::Error::custom)
                }),
            "title" => parts
                .pop_front()
                .map(Tag::Title)
                .ok_or_else(|| de::Error::custom("missing argument")),
            "t" => parts
                .pop_front()
                .map(Tag::Hashtag)
                .ok_or_else(|| de::Error::custom("missing argument")),
            "i" => {
                let external_id = parts
                    .pop_front()
                    .ok_or_else::<D::Error, _>(|| de::Error::custom("missing external id"));
                let url = parts
                    .pop_front()
                    .map(|url| url.parse::<Url>().map_err::<D::Error, _>(de::Error::custom))
                    .transpose();

                if let Ok(external_id) = external_id {
                    url.map(|url| Tag::ExternalContentId(external_id, url))
                } else {
                    Err(de::Error::custom("invalid external id"))
                }
            }
            "r" => {
                let url = parts
                    .pop_front()
                    .ok_or_else::<D::Error, _>(|| de::Error::custom("missing url"))
                    .and_then(|url| url.parse().map_err(de::Error::custom));

                let access = parts
                    .pop_front()
                    .map(|x| match x.as_str() {
                        "read" => Ok(RelayAccessType::Read),
                        "write" => Ok(RelayAccessType::Write),
                        _ => Err(de::Error::custom("invalid relay access type")),
                    })
                    .unwrap_or(Ok(RelayAccessType::Both));

                if let Ok(url) = url {
                    access.map(|access| Tag::Relay(url, access))
                } else {
                    Err(de::Error::custom("invalid url"))
                }
            }
            _ => Ok(default.clone()),
        };

        if !parts.is_empty() {
            return Ok(default);
        }

        Ok(tag.unwrap_or(default))
    }
}

#[cfg(test)]
mod test {
    use serde_json::json;

    use super::*;
    use crate::Response;

    #[test]
    fn bug_01() {
        let json = r#"["EVENT","d45a98f898820258a3313f5cb14f5fe8a9263437931ac6309f23ae0324833f39",{"content":"Will fight under the light of lightnings! No darkness on nostr! 🐶🐾😂","created_at":1678685477,"id":"5462822032c16d32267ba40536409fd51ea188b20e7dd5f9e7a0aa5561346f79","kind":1,"pubkey":"8fb140b4e8ddef97ce4b821d247278a1a4353362623f64021484b372f948000c","sig":"62c8e09a0fedd8096a313ef12436b0f0bcad56e9058d4bc12f61ae6094c099bb966dce47efa2d721c5bfdf923614db46d1d1c248105cb99c4ec495292cc875b1","tags":[["e","5f63f9e7d37673e76ddf7448cd67c3d74f9be96c240a40199b59a30db32d7f43"],["e","c70894331986c66a2baf6fc12dd5c86280e4616cee0e57bbee90972ebbb4b735"],["p","ac3fb436a663b25893657f4b6a3d9d2f02d1974bb5ced603f4d0c8ee32d7e0a2"]]}]"#;
        let message: Result<Response, _> = serde_json::from_str(json);
        assert!(message.is_ok());
    }

    #[test]
    fn parse_unexpected_tag() {
        let json = r#"["EVENT","f4e62282fda9c7d93a6e3b03fd1f1a3a34936f74584b0e021edf9659fd7da9d6",{"content":"1678682368","created_at":1678682368,"id":"7506ccd8ce4de835c61c04fd16f8489b2acb8cc052ed6730cba203188dfedf57","kind":30000,"pubkey":"228cc1e37a8fec2eee3dda3a3dbd04a60968086d8f42751a7632499d938eda8f","sig":"73ec91ee5d23a93287700f31274a57d84f46c0aa06523138c8f5b0cf2f20bd8a7db72346ecae9c717d9642cea81d72c35ea95ec615146b1640f62de5e3fbbb69","tags":[["d","chats/null/lastOpened"]]}]"#;
        let message: Result<Response, _> = serde_json::from_str(json);
        assert!(message.is_ok());
    }

    #[test]
    fn serialize_deserialize() {
        let json = json!(["p", "a0b0c0d0"]);
        assert_eq!(
            Tag::PubKey(PubKey {
                id: Addr::try_from_public_key_str("a0b0c0d0", None).expect("valid addr"),
                relayer_url: None,
                pet_name: None,
            }),
            serde_json::from_value(json).expect("valid json"),
        );
    }

    #[test]
    fn test_relay() {
        let json = json!(["r", "https://example.com", "read"]);
        assert_eq!(
            Tag::Relay(
                "https://example.com".parse().expect("valid url"),
                RelayAccessType::Read
            ),
            serde_json::from_value(json).expect("valid json"),
        );

        let json = json!(["r", "https://example.com", "write"]);
        assert_eq!(
            Tag::Relay(
                "https://example.com".parse().expect("valid url"),
                RelayAccessType::Write
            ),
            serde_json::from_value(json).expect("valid json"),
        );

        let json = json!(["r", "https://example.com"]);
        assert_eq!(
            Tag::Relay(
                "https://example.com".parse().expect("valid url"),
                RelayAccessType::Both,
            ),
            serde_json::from_value(json).expect("valid json"),
        );
    }

    #[test]
    fn test_relay_invalid_url() {
        let json = json!(["r", "example.com", "read"]);
        assert_eq!(
            Tag::Unknown(
                "r".to_string(),
                vec!["example.com".to_string(), "read".to_string()],
            ),
            serde_json::from_value(json).expect("valid json"),
        );
    }

    #[test]
    fn malformed_tag() {
        let json = vec![
            json!(["r"]),
            json!(["t"]),
            json!(["p"]),
            json!(["e"]),
            json!(["i"]),
            json!(["title"]),
        ];

        for json in json {
            let tag: Tag = serde_json::from_value(json).expect("valid");
            assert!(matches!(tag, Tag::Unknown(_, _)));
        }
    }

    #[test]
    fn test_relay_invalid_access() {
        let json = json!(["r", "https://example.com", "invalid"]);
        assert_eq!(
            Tag::Unknown(
                "r".to_string(),
                vec!["https://example.com".to_string(), "invalid".to_string()],
            ),
            serde_json::from_value(json).expect("valid json"),
        );
    }

    #[test]
    fn hashtag() {
        let json = json!(["t", "rust"]);
        assert_eq!(
            Tag::Hashtag("rust".to_string()),
            serde_json::from_value(json).expect("valid json"),
        );
    }

    #[test]
    fn title() {
        let json = json!(["title", "Rust"]);
        assert_eq!(
            Tag::Title("Rust".to_string()),
            serde_json::from_value(json).expect("valid json"),
        );
    }

    #[test]
    fn external_content_id() {
        let json = json!(["i", "123"]);
        assert_eq!(
            serde_json::from_value::<Tag>(json).expect("valid"),
            Tag::ExternalContentId("123".to_string(), None,),
        );
        let json = json!(["i", "123", "https://example.com"]);
        assert_eq!(
            serde_json::from_value::<Tag>(json).expect("valid"),
            Tag::ExternalContentId(
                "123".to_string(),
                Some("https://example.com".parse().expect("valid url"))
            ),
        );
    }

    #[test]
    fn external_content_id_invalid_args() {
        let json = json!(["i", "123", "https://example.com", "unexpected"]);
        assert_eq!(
            serde_json::from_value::<Tag>(json).expect("valid"),
            Tag::Unknown(
                "i".to_owned(),
                vec![
                    "123".to_owned(),
                    "https://example.com".to_owned(),
                    "unexpected".to_owned()
                ]
            ),
        );
    }

    #[test]
    fn external_content_id_invalid_url() {
        let json = json!(["i", "123", "facebook.com"]);
        assert_eq!(
            serde_json::from_value::<Tag>(json).expect("valid"),
            Tag::Unknown(
                "i".to_owned(),
                vec!["123".to_owned(), "facebook.com".to_owned()]
            ),
        );
    }
}
